%global forgeurl https://github.com/max0x7ba/atomic_queue
%global commit aa08199a7a516a561be1685afb644cf99e5b98e9

Name:           atomic-queue
# Upstream has never versioned the library.
Version:        0
%forgemeta
Release:        0.1%{?dist}
Summary:        C++ lockless queue

License:        MIT
URL:            %{forgeurl}
Source0:        %{forgesource}

# The library uses mandatory architecture-specific intrinsics or assembly for
# spin_loop_pause() in include/atomic_queue/defs.h. Upstream only provides
# implementations for these architectures:
ExclusiveArch:  x86_64 %{ix86} %{arm32} %{arm64}

BuildRequires:  gcc-c++
BuildRequires:  meson
BuildRequires:  boost-devel

# No compiled binaries are installed, so this would be empty.
%global debug_package %{nil}

%global common_description %{expand:
C++14 multiple-producer-multiple-consumer lockless queues based on circular
buffer with std::atomic.

The main design principle these queues follow is minimalism: the bare minimum
of atomic operations, fixed size buffer, value semantics.

These qualities are also limitations:

  • The maximum queue size must be set at compile time or construction time.
    The circular buffer side-steps the memory reclamation problem inherent in
    linked-list based queues for the price of fixed buffer size. See Effective
    memory reclamation for lock-free data structures in C++ for more details.
    Fixed buffer size may not be that much of a limitation, since once the
    queue gets larger than the maximum expected size that indicates a problem
    that elements aren’t processed fast enough, and if the queue keeps growing
    it may eventually consume all available memory which may affect the entire
    system, rather than the problematic process only. The only apparent
    inconvenience is that one has to do an upfront back-of-the-envelope
    calculation on what would be the largest expected/acceptable queue size.
  • There are no OS-blocking push/pop functions. This queue is designed for
    ultra-low-latency scenarios and using an OS blocking primitive would be
    sacrificing push-to-pop latency. For lowest possible latency one cannot
    afford blocking in the OS kernel because the wake-up latency of a blocked
    thread is about 1-3 microseconds, whereas this queue’s round-trip time can
    be as low as 150 nanoseconds.

Ultra-low-latency applications need just that and nothing more. The minimalism
pays off, see the throughput and latency benchmarks.

Available containers are:

  • AtomicQueue - a fixed size ring-buffer for atomic elements.
  • OptimistAtomicQueue - a faster fixed size ring-buffer for atomic elements
    which busy-waits when empty or full.
  • AtomicQueue2 - a fixed size ring-buffer for non-atomic elements.
  • OptimistAtomicQueue2 - a faster fixed size ring-buffer for non-atomic
    elements which busy-waits when empty or full.

These containers have corresponding AtomicQueueB, OptimistAtomicQueueB,
AtomicQueueB2, OptimistAtomicQueueB2 versions where the buffer size is
specified as an argument to the constructor.

Totally ordered mode is supported. In this mode consumers receive messages in
the same FIFO order the messages were posted. This mode is supported for push
and pop functions, but for not the try_ versions. On Intel x86 the totally
ordered mode has 0 cost, as of 2019.

Single-producer-single-consumer mode is supported. In this mode, no
read-modify-write instructions are necessary, only the atomic loads and stores.
That improves queue throughput significantly.}

%description %{common_description}


%package devel
Summary:        Development files for %{name}

# Header-only library
Provides:       %{name}-static = %{version}-%{release}

%description devel %{common_description}

The %{name}-devel package contains libraries and header files for
developing applications that use %{name}.


%prep
%forgeautosetup


%build
# Adjust C++ standard to match the system default.
%meson -Dbenchmarks=false -Dcpp_std=gnu++17
%meson_build


%install
# Upstream does not install with meson. Just copy the headers.
install -d '%{buildroot}%{_includedir}'
cp -rvp include/atomic_queue '%{buildroot}%{_includedir}/'


%check
%meson_test


%files devel
%license LICENSE
%doc README.md
# We do NOT package “html”, which contains benchmark results, because it loads
# a Google Analytics script.

%{_includedir}/atomic_queue


%changelog
* Tue Aug 17 2021 Benjamin A. Beasley <code@musicinmybrain.net> - 0-0.1.20210816gitaa08199
- Initial package for EPEL8
